module.exports = grammar({
  name: 'MooLang',

  rules: {
    source_file: $ => repeat($._declaration),

    _declaration: $ => choice(
      $.function_declaration,
      seq($.variable_declaration, $.eol),
      $._statement
    ),

    function_declaration: $ => seq(
      'fun',
      $.identifier,
      $.function_literal
    ),

    variable_declaration: $ => seq(
      choice('var', 'let'),
      $.identifier,
      optional(seq(
        ':', $._type
      )),
      optional(seq(
        '=', $._expression
      ))
    ),

    _statement: $ => choice(
      $.while_statement,
      $.for_statement,
      $.when_statement,
      $.block_statement,
      $.return_statement,
      $.expression_statement
    ),

    when_statement: $ => prec.left(5, seq(
      'when',
      choice(seq(
          $.clause_result,
          optional($.else_result)
        ),
        seq(
          optional(seq(
            '(',
            $._expression,
            ')'
          )),
          '{',
          $.clause_list,
          optional($.else_result),
          '}',
        )
      ),
    )),

    for_statement: $ => prec(4, seq(
      'for',
      '(',
      $.variable_declaration,
      'in',
      $._expression,
      ')',
      $._statement
    )),

    while_statement: $ => prec(3, seq(
      'while',
      '(',
      $._expression,
      ')',
      $._statement
    )),

    block_statement: $ => prec(2, seq(
      '{',
      repeat($._declaration),
      '}'
    )),

    return_statement: $ => seq(
      'return',
      $._expression,
      repeat(seq(',', $._expression)),
      $.eol
    ),

    else_result: $ => prec(1, seq(
      'else',
      ':',
      $._statement
    )),

    clause_result: $ => prec(0, seq(
      $._expression,
      ':',
      $._statement
    )),

    clause_list: $ => repeat1($.clause_result),

    expression_statement: $ => prec(1, seq(
      $._expression,
      $.eol
    )),

    _expression: $ => choice(
      $.assignment,
      $.function_call,
      $.range_literal,
      $.comparison_expression,
      $.addition_expression,
      $.multiplication_expression,
      $.cast_expression,
      $.unary_expression,
      prec(5, $.primary),
    ),

    assignment: $ => prec.right(0, seq(
      $.identifier,
      '=',
      $._expression
    )),

    function_call: $ => prec(1, seq(
      $.identifier,
      '(',
      optional($.argument_list),
      ')'
    )),

    argument_list: $ => seq(
      $._expression,
      repeat(seq(
        ',',
        $._expression
      ))
    ),

    range_literal: $ => prec(2, seq(
      '[',
      $._expression,
      choice('...', '..<'),
      $._expression,
      ']'
    )),

    comparison_expression: $ => prec.left(3, seq(
      $._expression,
      choice('<', '<=', '>', '>=', '==', '!='),
      $._expression
    )),

    addition_expression: $ => prec.left(4, seq(
      $._expression,
      choice('+', '-'),
      $._expression
    )),

    multiplication_expression: $ => prec.left(5, seq(
      $._expression,
      choice('*', '/', '%'),
      $._expression
    )),

    cast_expression: $ => prec.left(6, seq(
      $._expression,
      'as',
      $._type
    )),

    unary_expression: $=> prec.right(7, seq(
      choice('!', '-'),
      $.primary
    )),

    _type: $ => choice(
      $.primitive_type,
      $.list_type,
      $.map_type
    ),

    primitive_type: $ => choice(
      'bool', 'string',
      'int8', 'int16', 'int32', 'int64',
      'uint8', 'uint16', 'uint32', 'uint64',
      'float32', 'float64'
    ),

    list_type: $ => seq(
      '[',
      $._type,
      ']'
    ),

    map_type: $ => seq(
      '{',
      $._type,
      ':',
      $._type,
      '}'
    ),

    primary: $ => choice(
      $.function_literal,
      $.bool_literal,
      $.list_literal,
      $.map_literal,
      $.parenthetical_expression,
      $.identifier,
      $.number
    ),

    bool_literal: $ => choice('true', 'false'),

    function_literal: $ => seq(
      '(',
      optional($.parameter_list),
      ')',
      optional(seq(':', $.type_list)),
      $.block_statement
    ),

    parameter_list: $ => seq(
      $.parameter,
      repeat(seq(',', $.parameter))
    ),

    parameter: $ => seq(
      $.identifier,
      ':',
      $._type
    ),

    type_list: $ => seq(
      $._type,
      repeat(seq(',', $._type))
    ),

    list_literal: $ => seq(
      '[',
      optional(seq(
        $._expression,
        repeat(seq(
          ',',
          $._expression
        )),
        optional(','),
      )),
      ']'
    ),

    map_literal: $ => seq(
      '{',
      optional(seq(
        $._expression,
        ':',
        $._expression,
        repeat(seq(
          ',',
          $._expression,
          ':',
          $._expression,
        )),
      )),
      '}'
    ),

    parenthetical_expression: $ => seq(
      '(',
      $._expression,
      ')'
    ),

    identifier: $ => prec(99, /[a-zA-Z_][a-zA-Z0-9_]*/),

    number: $ => prec(99, /\d+(_\d+)*(\.\d+)?/),

    eol: $ => '\n'
  }
});
